const { app, ipcMain, protocol, session, BrowserWindow, BrowserView, Menu, nativeImage, clipboard, dialog, Notification } = require('electron');
const path = require('path');
const fs = require('fs');
const url = require('url');
const os = require('os');

const localShortcut = require("electron-localshortcut");

const Config = require('electron-store');
const config = new Config({
    defaults: {
        design: {
            homeButton: false,
            darkTheme: false,
            theme: 'default'
        },
        homePage: {
            defaultPage: 'my://newtab',
            defaultEngine: 'Google',
            searchEngines: [
                {
                    name: 'Google',
                    url: 'https://www.google.com/search?q=%s'
                },
                {
                    name: 'Bing',
                    url: 'https://www.bing.com/search?q=%s'
                },
                {
                    name: 'Yahoo! Japan',
                    url: 'https://search.yahoo.co.jp/search?p=%s'
                },
                {
                    name: 'goo',
                    url: 'https://search.goo.ne.jp/web.jsp?MT=%s'
                },
                {
                    name: 'Google Translate',
                    url: 'https://translate.google.com/?text=%s'
                },
                {
                    name: 'Youtube',
                    url: 'https://www.youtube.com/results?search_query=%s'
                },
                {
                    name: 'Twitter',
                    url: 'https://www.twitter.com/search?q=%s'
                },
                {
                    name: 'GitHub',
                    url: 'https://github.com/search?q=%s'
                }
            ]
        },
        adBlocker: true,
        window: {
            isCustomTitlebar: true,
            isMaximized: false,
            bounds: {
                width: 1100,
                height: 680
            }
        }
    },
});

const Datastore = require('nedb');
let db = {};

db.history = new Datastore({
    filename: path.join(app.getPath('userData'), 'Files', 'History.db'),
    autoload: true,
    timestampData: true
});

const { loadFilters, updateFilters, removeAds } = require('./AdBlocker');

let views = [];
let tabCount = 0;

getBaseWindow = (width = 1100, height = 680, minWidth = 320, minHeight = 600, x, y, frame = false) => {
    return new BrowserWindow({
        width, height, minWidth, minHeight, x, y, 'titleBarStyle': 'hidden', frame, fullscreenable: true,
        webPreferences: {
            nodeIntegration: true,
            webviewTag: true,
            plugins: true,
            experimentalFeatures: true,
            contextIsolation: false,
        }
    });
}

registerProtocols = () => {
    protocol.registerFileProtocol('my', (request, callback) => {
        const parsed = url.parse(request.url);

        if (parsed.hostname.endsWith('.css') || parsed.hostname.endsWith('.js')) {
            return callback({
                path: path.join(app.getAppPath(), 'pages', parsed.hostname),
            });
        } else {
            return callback({
                path: path.join(app.getAppPath(), 'pages', `${parsed.hostname}.html`),
            });
        }
    }, (error) => {
        if (error) console.error('Failed to register protocol: ' + error);
    });

    session.fromPartition('persist:private').protocol.registerFileProtocol('my', (request, callback) => {
        const parsed = url.parse(request.url);

        if (parsed.hostname.endsWith('.css') || parsed.hostname.endsWith('.js')) {
            return callback({
                path: path.join(app.getAppPath(), 'pages', parsed.hostname),
            });
        } else {
            return callback({
                path: path.join(app.getAppPath(), 'pages', `${parsed.hostname}.html`),
            });
        }
    }, (error) => {
        if (error) console.error('Failed to register protocol: ' + error);
    });
}

registerProtocolWithPrivateMode = (windowId) => {
    session.fromPartition(`persist:${windowId}`).protocol.registerFileProtocol('my', (request, callback) => {
        const parsed = url.parse(request.url);

        if (parsed.hostname.endsWith('.css') || parsed.hostname.endsWith('.js')) {
            return callback({
                path: path.join(app.getAppPath(), 'pages', parsed.hostname),
            });
        } else {
            return callback({
                path: path.join(app.getAppPath(), 'pages', `${parsed.hostname}.html`),
            });
        }
    }, (error) => {
        if (error) console.error('Failed to register protocol: ' + error);
    });
}

module.exports = class WindowManager {
    constructor() {
        this.windows = new Map();

        console.log('WindowManager: constructor');

        ipcMain.on('window-add', (e, args) => {
            this.addWindow(args.isPrivate);
        });
    }

    addWindow = (isPrivate = false) => {
        loadFilters();

        const { width, height, x, y } = config.get('window.bounds');
        const window = getBaseWindow(config.get('window.isMaximized') ? 1110 : width, config.get('window.isMaximized') ? 680 : height, 320, 600, x, y, !config.get('window.isCustomTitlebar'));

        const id = (!isPrivate ? window.id : `private-${window.id}`);

        config.get('window.isMaximized') && window.maximize();

        const startUrl = process.env.ELECTRON_START_URL || url.format({
            pathname: path.join(__dirname, '/../build/index.html'), // 警告：このファイルを移動する場合ここの相対パスの指定に注意してください
            protocol: 'file:',
            slashes: true,
            hash: `/${id}`,
        });

        window.loadURL(startUrl);

        localShortcut.register(window, 'CmdOrCtrl+Shift+I', () => {
            if (window.getBrowserView() == undefined) return;

            const view = window.getBrowserView();
            if (view.webContents.isDevToolsOpened()) {
                view.webContents.devToolsWebContents.focus();
            } else {
                view.webContents.openDevTools();
            }
        });

        localShortcut.register(window, 'CmdOrCtrl+R', () => {
            if (window.getBrowserView() == undefined) return;

            const view = window.getBrowserView();
            view.webContents.reload();
        });

        localShortcut.register(window, 'CmdOrCtrl+Shift+R', () => {
            if (window.getBrowserView() == undefined) return;

            const view = window.getBrowserView();
            view.webContents.reloadIgnoringCache();
        });

        window.on('closed', () => {
            this.windows.delete(id);
        });

        ['resize', 'move'].forEach(ev => {
            window.on(ev, () => {
                config.set('window.isMaximized', window.isMaximized());
                config.set('window.bounds', window.getBounds());
            })
        });

        window.on('close', (e) => {
            for (var i = 0; i < views.length; i++) {
                if (views[i].windowId == id) {
                    views.splice(i, 1);
                }
            }
        });

        window.on('maximize', this.fixBounds.bind(this, id));
        window.on('unmaximize', this.fixBounds.bind(this, id));
        window.on('enter-full-screen', this.fixBounds.bind(this, id));
        window.on('leave-full-screen', this.fixBounds.bind(this, id));
        window.on('enter-html-full-screen', this.fixBounds.bind(this, id));
        window.on('leave-html-full-screen', this.fixBounds.bind(this, id));

        // registerProtocols();
        this.registerListeners(id);

        this.windows.set(id, window);
    }

    registerListeners = (id) => {
        ipcMain.on(`browserview-add-${id}`, (e, args) => {
            this.addView(id, args.url, args.isActive);
        });

        ipcMain.on(`browserview-remove-${id}`, (e, args) => {
            this.removeView(id, args.id);
        });

        ipcMain.on(`browserview-select-${id}`, (e, args) => {
            this.selectView(id, args.id);
        });

        ipcMain.on(`browserview-get-${id}`, (e, args) => {
            let datas = [];
            for (var i = 0; i < views.length; i++) {
                if (views[i].windowId == id) {
                    const url = views[i].view.webContents.getURL();
                    datas.push({ id: views[i].id, title: views[i].view.webContents.getTitle(), url: url, icon: url.startsWith('my://') ? undefined : `http://www.google.com/s2/favicons?domain=${url}` });
                }
            }
            e.sender.send(`browserview-get-${id}`, { views: datas });
        });

        ipcMain.on(`browserview-goBack-${id}`, (e, args) => {
            views.filter(function (view, i) {
                if (view.id == args.id) {
                    let webContents = views[i].view.webContents;
                    if (webContents.canGoBack())
                        webContents.goBack();
                }
            });
        });

        ipcMain.on(`browserview-goForward-${id}`, (e, args) => {
            views.filter(function (view, i) {
                if (view.id == args.id) {
                    let webContents = views[i].view.webContents;
                    if (webContents.canGoForward())
                        webContents.goForward();
                }
            });
        });

        ipcMain.on(`browserview-reload-${id}`, (e, args) => {
            views.filter(function (view, i) {
                if (view.id == args.id) {
                    let webContents = views[i].view.webContents;
                    webContents.reload();
                }
            });
        });

        ipcMain.on(`browserview-stop-${id}`, (e, args) => {
            views.filter(function (view, i) {
                if (view.id == args.id) {
                    let webContents = views[i].view.webContents;
                    webContents.stop();
                }
            });
        });

        ipcMain.on(`browserview-goHome-${id}`, (e, args) => {
            views.filter(function (view, i) {
                if (view.id == args.id) {
                    let webContents = views[i].view.webContents;
                    webContents.loadURL(config.get('homePage.defaultPage'));
                }
            });
        });

        ipcMain.on(`browserview-loadURL-${id}`, (e, args) => {
            views.filter(function (view, i) {
                if (view.id == args.id) {
                    let webContents = views[i].view.webContents;
                    webContents.loadURL(args.url);
                }
            });
        });

        ipcMain.on(`browserview-loadFile-${id}`, (e, args) => {
            views.filter(function (view, i) {
                if (view.id == args.id) {
                    let webContents = views[i].view.webContents;
                    webContents.loadFile(args.url);
                }
            });
        });
    }

    updateNavigationState = (windowId, id, view) => {
        const window = this.windows.get(windowId);
        window.webContents.send(`update-navigation-state-${windowId}`, {
            id: id,
            canGoBack: view.webContents.canGoBack(),
            canGoForward: view.webContents.canGoForward(),
        });
    }

    fixBounds = (windowId) => {
        const window = this.windows.get(windowId);
        if (window.getBrowserView() == undefined) return;

        const view = window.getBrowserView();
        const { width, height } = window.getContentBounds();

        view.setAutoResize({ width: true, height: true });
        if (window.isFullScreen()) {
            view.setBounds({
                x: 0,
                y: 0,
                width: width,
                height: height,
            });
        } else {
            view.setBounds({
                x: 1,
                y: 73 + 1,
                width: width - 2,
                height: window.isMaximized() ? height - 73 : (height - 73) - 2,
            });
        }
        view.setAutoResize({ width: true, height: true });
    }

    addView = (windowId, url, isActive) => {
        const id = tabCount++;
        this.addTab(windowId, id, url, isActive);
    }

    removeView = (windowId, id) => {
        views.filter((view, i) => {
            if (windowId == view.windowId && id == view.id) {
                views.splice(i, 1);
            }
        });
    }

    selectView = (windowId, id) => {
        const window = this.windows.get(windowId);
        views.filter((view, i) => {
            if (windowId == view.windowId && id == view.id) {
                window.setBrowserView(views[i].view);
                window.setTitle(views[i].view.webContents.getTitle());
                window.webContents.send(`browserview-set-${windowId}`, { id: id });
                this.fixBounds(windowId);
            }
        });
    }

    getViews = (windowId) => {
        let datas = [];
        for (var i = 0; i < views.length; i++) {
            if (views[i].windowId == windowId) {
                const url = views[i].view.webContents.getURL();
                datas.push({ id: views[i].id, title: views[i].view.webContents.getTitle(), url: url, icon: url.startsWith('my://') ? undefined : `http://www.google.com/s2/favicons?domain=${url}` });
            }
        }
        const window = this.windows.get(windowId);
        window.webContents.send(`browserview-get-${windowId}`, { views: datas });
    }

    addTab = (windowId, id, url = config.get('homePage.defaultPage'), isActive = true) => {
        const window = this.windows.get(windowId);

        const view = new BrowserView({
            webPreferences: {
                nodeIntegration: false,
                contextIsolation: false,
                plugins: true,
                experimentalFeatures: true,
                safeDialogs: true,
                safeDialogsMessage: '今後このページではダイアログを表示しない',
                ...(String(windowId).startsWith('private') && { partition: `persist:${windowId}` }),
                preload: require.resolve('./Preload')
            }
        });

        if (String(windowId).startsWith('private')) {
            registerProtocolWithPrivateMode(windowId);
        }

        view.webContents.on('did-start-loading', () => {
            window.webContents.send(`browserview-start-loading-${windowId}`, { id: id });
        });
        view.webContents.on('did-stop-loading', () => {
            window.webContents.send(`browserview-stop-loading-${windowId}`, { id: id });
        });

        view.webContents.on('did-finish-load', (e) => {
            window.setTitle(view.webContents.getTitle());
            window.webContents.send(`browserview-load-${windowId}`, { id: id, title: view.webContents.getTitle(), url: view.webContents.getURL() });

            this.updateNavigationState(windowId, id, view);
        });

        view.webContents.on('did-start-navigation', (e) => {
            const url = view.webContents.getURL();

            if (config.get('adBlocker'))
                removeAds(url, view.webContents);

            this.updateNavigationState(windowId, id, view);
        });

        view.webContents.on('page-title-updated', (e) => {
            window.setTitle(view.webContents.getTitle());
            window.webContents.send(`browserview-load-${windowId}`, { id: id, title: view.webContents.getTitle(), url: view.webContents.getURL() });

            if (!String(windowId).startsWith('private') && !view.webContents.getURL().startsWith('my://'))
                db.history.insert({ title: view.webContents.getTitle(), url: view.webContents.getURL() });

            this.updateNavigationState(windowId, id, view);
        });

        view.webContents.on('page-favicon-updated', (e, favicons) => {
            window.setTitle(view.webContents.getTitle());
            if (favicons.length > 0) {
                window.webContents.send(`browserview-load-${windowId}`, { id: id, title: view.webContents.getTitle(), url: view.webContents.getURL(), favicon: favicons[0] });
            } else {
                window.webContents.send(`browserview-load-${windowId}`, { id: id, title: view.webContents.getTitle(), url: view.webContents.getURL(), favicon: favicons[0] });
            }

            this.updateNavigationState(windowId, id, view);
        });

        view.webContents.on('did-change-theme-color', (e, color) => {
            window.setTitle(view.webContents.getTitle());
            window.webContents.send(`browserview-load-${windowId}`, { id: id, title: view.webContents.getTitle(), url: view.webContents.getURL(), color: color });
        });

        view.webContents.on('new-window', (e, url) => {
            e.preventDefault();
            this.addView(windowId, url, true);
        });

        view.webContents.on('context-menu', (e, params) => {
            const menu = Menu.buildFromTemplate(
                [
                    ...(params.linkURL !== '' ?
                        [
                            {
                                label: '新しいタブで開く',
                                click: () => {
                                    this.addView(windowId, params.linkURL, true);
                                }
                            },
                            {
                                label: '新しいウィンドウで開く',
                                enabled: false,
                                click: () => { view.webContents.openDevTools(); }
                            },
                            {
                                label: 'シークレットウィンドウで開く',
                                enabled: false,
                                click: () => { view.webContents.openDevTools(); }
                            },
                            { type: 'separator' },
                            {
                                label: 'リンクをコピー',
                                accelerator: 'CmdOrCtrl+C',
                                click: () => {
                                    clipboard.clear();
                                    clipboard.writeText(params.linkURL);
                                }
                            },
                            { type: 'separator' }
                        ] : []),
                    ...(params.hasImageContents ?
                        [
                            {
                                label: '新しいタブで画像を開く',
                                click: () => {
                                    this.addView(windowId, params.srcURL, true);
                                }
                            },
                            {
                                label: '画像をコピー',
                                click: () => {
                                    const img = nativeImage.createFromDataURL(params.srcURL);

                                    clipboard.clear();
                                    clipboard.writeImage(img);
                                }
                            },
                            {
                                label: '画像アドレスをコピー',
                                click: () => {
                                    clipboard.clear();
                                    clipboard.writeText(params.srcURL);
                                }
                            },
                            { type: 'separator' }
                        ] : []),
                    ...(params.isEditable ?
                        [
                            {
                                label: '元に戻す',
                                accelerator: 'CmdOrCtrl+Z',
                                enabled: params.editFlags.canUndo,
                                click: () => { view.webContents.undo(); }
                            },
                            {
                                label: 'やり直す',
                                accelerator: 'CmdOrCtrl+Y',
                                enabled: params.editFlags.canRedo,
                                click: () => { view.webContents.redo(); }
                            },
                            { type: 'separator' },
                            {
                                label: '切り取り',
                                accelerator: 'CmdOrCtrl+X',
                                enabled: params.editFlags.canCut,
                                click: () => { view.webContents.cut(); }
                            },
                            {
                                label: 'コピー',
                                accelerator: 'CmdOrCtrl+C',
                                enabled: params.editFlags.canCopy,
                                click: () => { view.webContents.copy(); }
                            },
                            {
                                label: '貼り付け',
                                accelerator: 'CmdOrCtrl+V',
                                enabled: params.editFlags.canPaste,
                                click: () => { view.webContents.paste(); }
                            },
                            { type: 'separator' },
                            {
                                label: 'すべて選択',
                                accelerator: 'CmdOrCtrl+A',
                                enabled: params.editFlags.canSelectAll,
                                click: () => { view.webContents.selectAll(); }
                            },
                            { type: 'separator' }
                        ] : []),
                    ...(params.selectionText !== '' && !params.isEditable ?
                        [
                            {
                                label: 'コピー',
                                accelerator: 'CmdOrCtrl+C',
                                click: () => { view.webContents.copy(); }
                            },
                            {
                                label: `Googleで「${params.selectionText}」を検索`,
                                click: () => {
                                    this.addView(windowId, `https://www.google.co.jp/search?q=${params.selectionText}`, true);
                                }
                            },
                            { type: 'separator' }
                        ] : []),
                    {
                        label: '戻る',
                        accelerator: 'Alt+Left',
                        enabled: view.webContents.canGoBack(),
                        click: () => { view.webContents.goBack(); }
                    },
                    {
                        label: '進む',
                        accelerator: 'Alt+Right',
                        enabled: view.webContents.canGoForward(),
                        click: () => { view.webContents.goForward(); }
                    },
                    {
                        label: '再読み込み',
                        accelerator: 'CmdOrCtrl+R',
                        click: () => { view.webContents.reload(); }
                    },
                    { type: 'separator' },
                    {
                        label: 'ページの保存',
                        accelerator: 'CmdOrCtrl+S',
                        enabled: !view.webContents.getURL().startsWith('my://'),
                        click: () => {
                            view.webContents.savePage(`${app.getPath('downloads')}/${view.webContents.getTitle()}.html`, 'HTMLComplete', (err) => {
                                if (!err) console.log('Page Save successfully');
                            });
                        }
                    },
                    {
                        label: '印刷',
                        accelerator: 'CmdOrCtrl+P',
                        enabled: !view.webContents.getURL().startsWith('my://'),
                        click: () => { view.webContents.print(); }
                    },
                    { type: 'separator' },
                    {
                        label: 'デベロッパーツール',
                        accelerator: 'CmdOrCtrl+Shift+I',
                        enabled: !view.webContents.getURL().startsWith('my://'),
                        click: () => { if (view.webContents.isDevToolsOpened()) { view.webContents.devToolsWebContents.focus(); } else { view.webContents.openDevTools(); } }
                    }
                ]
            );

            menu.popup();
        });

        view.webContents.on('before-input-event', (e, input) => {
            if ((input.control || input.meta) && input.shift && input.key == 'I') {
                e.preventDefault();
                if (view.webContents.isDevToolsOpened()) {
                    view.webContents.devToolsWebContents.focus();
                } else {
                    view.webContents.openDevTools();
                }
            } else if ((input.control || input.meta) && input.key == 'R') {
                e.preventDefault();
                view.webContents.reload();
            } else if ((input.control || input.meta) && input.shift && input.key == 'R') {
                e.preventDefault();
                view.webContents.reloadIgnoringCache();
            }
        });

        view.webContents.session.on('will-download', (event, item, webContents) => {
            item.on('updated', (e, state) => {
                if (state === 'interrupted') {
                    console.log('Download is interrupted but can be resumed')
                } else if (state === 'progressing') {
                    if (item.isPaused()) {
                        console.log('Download is paused')
                    } else {
                        console.log(`Received bytes: ${item.getReceivedBytes()}`)
                    }
                }
            });

            item.once('done', (e, state) => {
                if (state === 'completed') {
                    console.log('Download successfully')
                } else {
                    console.log(`Download failed: ${state}`)
                }
            });
        });

        view.webContents.loadURL(url);

        views.push({ windowId, id, view });
        console.log(views);

        if (isActive) {
            window.webContents.send(`browserview-set-${windowId}`, { id: id });
            window.setBrowserView(view);
        }
        this.fixBounds(windowId);

        this.getViews(windowId);
    }
}